#include "lexer.h"
#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#define cpy() {\
    long siz = mCurr-mHelper;\
    if(siz <= SmallBuffSize) {\
        mBuff = mSmallBuff;\
    } else {\
        if(mBigBuffSize < siz) {\
            mBigBuff = (char*)realloc((char*)mBigBuff, siz+1);\
            mBigBuffSize = siz;\
        }\
        mBuff = mBigBuff;\
    }\
    memcpy(mBuff, mHelper, siz);\
    mBuff[siz] = 0;\
}

#define chr(c,tok) \
    if(*mCurr == c) {\
        mCurr++;\
        mBuff[0] = c; mBuff[1] = 0;\
        return tok;\
    }

#define ret(tok) cpy(); return tok;

#define str_(s,tok) {\
    char* tmp = s;\
    for(long i = 0; 1; i++) {\
        if(!tmp[i]) {\
            if(!isalnum(*mCurr)) {while(*mCurr != '\n' && *mCurr) mCurr++;ret(tok);}\
            else break;\
        }\
        if(tmp[i] != *mCurr) break;\
        if(!*mCurr) break;\
        mCurr++;\
    }\
    mCurr = mHelper;\
}

#define str(s,tok) {\
    char* tmp = s;\
    for(long i = 0; 1; i++) {\
        if(!tmp[i]) {\
            if(!isalnum(*mCurr)) {ret(tok);}\
            else break;\
        }\
        if(tmp[i] != *mCurr) break;\
        if(!*mCurr) break;\
        mCurr++;\
    }\
    mCurr = mHelper;\
}

#define op_1(c,pr,tok) \
    if(*mCurr == c) {\
        mCurr++;\
        mPr = pr;\
        mBuff[0] = c; mBuff[1] = 0;\
        return tok;\
    }

#define op_2(c1,pr1,tok1,c2,pr2,tok2) \
    if(*mCurr == c1) {\
        mCurr++;\
        mBuff[0] = c1;\
        if(*mCurr == c2) {\
            mCurr++;\
            mPr = pr2;\
            mBuff[1] = c2; mBuff[2] = 0;\
            return tok2;\
        }\
        mPr = pr1;\
        mBuff[1] = 0;\
        return tok1;\
}

Lexer::Lexer(char* pSource) {
    mSource = pSource;
    mCurr = pSource;
    mBuff = mSmallBuff;
    mEnd = false;
    mBigBuffSize = 0;
}

Lexer::~Lexer() {
    if(mBigBuffSize) free(mBigBuff);
}

const char* Lexer::ident() {
    return mBuff;
}

Lexer::Token Lexer::next() {
    while(*mCurr == ' ' || *mCurr == '\n' || *mCurr == '\t') mCurr++;
    if(*mCurr=='/' && *(mCurr+1)=='/') {
        mCurr=mCurr+2;
        while(*mCurr != '\n' && *mCurr) mCurr++;
        if(!*mCurr) return kToken__End;
        mCurr+=2;
        return next();
    }
    if(*mCurr=='/' && *(mCurr+1)=='*') {
        mCurr=mCurr+2;
        while(!(*mCurr == '*' && *(mCurr+1) == '/') && *mCurr) mCurr++;
        if(!*mCurr) return kToken__End;
        mCurr+=2;
        return next();
    }
    if(*mCurr == 0) {
        mEnd = false;
        return kToken__Invalid;
    }


    //simple symbols
    chr(';', kToken_SCol);
    chr(',', kToken_Comma);
    chr(':', kToken_Col);
    chr('#', kToken_Hash);
    chr('(', kToken_LParen);
    chr(')', kToken_RParen);
    chr('[', kToken_LBracket);
    chr(']', kToken_RBracket);
    chr('{', kToken_LBrace);
    chr('}', kToken_RBrace);

    mHelper = mCurr;

    //keywords
    //...
    //types aren't keywords, but identifiers

    str("class", kToken_Class);
    str("private", kToken_ShwMod);
    str("public", kToken_ShwMod);
    str("protected", kToken_ShwMod);
    //str("const", kToken_Const);
    //str("typedef", kToken_Typedef);
    //str("virtual", kToken_Virtual);
    //str("friend", kToken_Friend);
    str_("ifndef", kToken_Prag);
    str_("ifdef", kToken_Prag);
    str("endif", kToken_Prag);
    str_("define", kToken_Prag);

    //constants
    if(isdigit(*mCurr)) {
        mCurr++;
        while(isdigit(*mCurr)) *mCurr++;
        ret(kToken_Int);
    }

    str("false", kToken_False);
    str("true", kToken_True);

    //identifiers
    if(isalpha(*mCurr) || *mCurr == '_') {
        mCurr++;
        while(isalnum(*mCurr) || *mCurr == '_') *mCurr++;
        ret(kToken_Ident);
    }

    //operators
    chr('=', kToken_Binary);

    mBuff[0] = *mCurr;
    mBuff[1] = 0;
    mCurr++;
    return kToken__Invalid;
}
